import os
import re
from migranto.lib.db import Db
from migranto.lib.error import APIError
from migranto.lib.utils import getFileListByMask
from migranto.lib import sql as SQL
import sys

class Migranto:

    def __init__(self, options):
        self.path = options.path
        self.dbstring = options.database
        self.name = options.name
        self.toStep = options.migration if options.migration is not None else 'max'
        self.nowStep = None
        self.topStep = 0
        self.wasStep = None
        self.verbose = options.verbose
        self.migrations = {}
        self.storage = options.storage
        self.fake = options.fake
        self.home_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), '..')

    def dbConnect(self):
        self.db = Db(self.dbstring)
        self.SQL = getattr(SQL, self.db.scheme)

    def dbClose(self):
        self.db.close()

    def getLastMigration(self):
        self.nowStep = self.db.fetchOneResult(self.SQL['selectLatestMigration'].format(tablename = self.storage) , (self.name,))
        return self.nowStep

    def createNewMigration(self, name):
        try:
            self.db.begin()
            self.db.execute(self.SQL['insertMigration'].format(tablename = self.storage), (name,))
            self.db.commit()
        except Exception:
            self.db.rollback()
            self.db.execute(self.SQL['createTable'].format(tablename = self.storage))
            self.db.execute(self.SQL['insertMigration'].format(tablename = self.storage), (name,))
            self.db.commit()

    def loadMigrations(self):
        migration_files = getFileListByMask(self.path, ['*.sql'])

        r = re.compile('.*(\d+)_.*?(_rollback)?\.sql$', re.IGNORECASE)

        if migration_files:
            for migration_file in migration_files:
                m = r.findall(migration_file)[0]

                if m:
                    m_step = int(m[0])

                    if m_step in self.migrations:
                        migration = self.migrations[m_step]
                    else:
                        self.migrations[m_step] = migration = {}

                    if m[1]:
                        if "down" not in migration:
                            migration["down"] = migration_file
                        else:
                            raise APIError("Migration duplicates at %d step", m_step)
                    else:
                        if "up" not in migration:
                            migration["up"] = migration_file
                        else:
                            raise APIError("Migration duplicates at %d step", m_step)

                    if m_step > self.topStep:
                        self.topStep = m_step
                else:
                    continue
        else:
            raise APIError('No migration files found')

        if self.toStep == 'max':
            self.toStep = self.topStep


    def applyMigrations(self):
        if self.toStep > self.topStep:
            raise APIError('Top migration is %d' % self.topStep)

        # db is up to date
        if self.toStep == self.nowStep:
            raise APIError("Database for %s is up to date. No migration applied" % self.name )

        # down
        if self.toStep < self.nowStep:
            for i in range(self.nowStep, self.toStep, -1):
                self.applyMigration(self.migrations[i], 'down', i)

            return  self.toStep
        # up
        elif self.toStep > self.nowStep:
            # no migrations applied yet
            if not self.nowStep:
                self.nowStep = 1

            for i in range(self.nowStep, self.toStep + 1):
                self.applyMigration(self.migrations[i], 'up', i)

            return  self.toStep

    def applyMigration(self, migration, flow, number):
        if self.verbose:
            print "%s %d: %s" % (self.name, number, migration[flow])

        try:
            self.db.begin()
            if not self.fake:
                self.db.executeFile(migration[flow])
            self.db.execute(self.SQL['updateMigration'].format(tablename = self.storage), ( number - 1 if flow == 'down' else number, self.name))
            self.db.commit()

        except Exception, exc:
            try:
                self.db.rollback()
            except Exception, rollback_exc:
                raise APIError('Error: ' + str(exc) + 'Rollback error: ' + str(rollback_exc))
            raise APIError('Error: ' + str(exc))

    def printMigrationStatus(self):
        try:
            if not self.name:
                self.db.execute(self.SQL['selectAllMigrationStatus'].format(tablename = self.storage))
            else:
                self.db.execute(self.SQL['selectOneMigrationStatus'].format(tablename = self.storage), (self.name,))
        except Exception, exc:
            raise APIError(str(exc))

        print('Migranto status')
        self.db.printData(sys.stdout)

    def run(self):

        if self.verbose:
            if self.fake:
                print "FAKE MIGRATION!"
            print "\nOpening database connection...\n"
        self.dbConnect()

        if self.path and self.name:


            if self.verbose:
                print "Database: " + self.db.dbname

            self.loadMigrations()

            if self.getLastMigration() is None:
                self.createNewMigration(self.name)
                self.nowStep = 0

            if self.verbose:
                print "Migration for %s. Now %d / To %d / Max %d" % (self.name, self.nowStep, self.toStep, self.topStep)

            self.wasStep = self.nowStep
            self.nowStep = self.applyMigrations()

            print "%s successfully migrated from %d to %d" % (self.name, self.wasStep, self.nowStep)

        else:
            self.printMigrationStatus()

        if self.verbose:
            print "\nClosing database connection"
        self.dbClose()

        print "\n"


import urlparse
from migranto.lib.error import APIError
from migranto.lib.printer import Printer

class Db:

    schemes = ('pgsql', 'sqlite')

    def __init__(self, db_url):
        self.options = self.parseDBString(db_url)

        try:
            if self.options.scheme == 'pgsql':
                import psycopg2
                import psycopg2.extensions
                psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
                psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)
                self.dbname = self.options.path[1:]
                self.conn = psycopg2.connect(database=self.dbname, user=self.options.username, password=self.options.password, host=self.options.hostname, port=self.options.port)

            elif self.options.scheme == 'sqlite':
                import sqlite3
                self.dbname = self.options.path
                self.conn = sqlite3.connect(self.dbname)

        except Exception, exc:
            raise APIError(exc)

        self.c = self.conn.cursor()

    @property
    def scheme(self):
        return self.options.scheme

    def parseDBString(self, db_string):
        r = urlparse.urlparse(db_string)

        if r.scheme in self.schemes:
            return r
        else:
            raise APIError('Provide a correct database connection string')

    def printData(self, out):
        printer = Printer(out, [20, 10, 20])
        printer.printRow(('Name', 'Migration', 'Date'))

        for row in self.c:
            printer.printRow(row)

    def fetchOneResult(self, sql, data):
        try:
            self.c.execute(sql, data)
        except Exception:
            return None

        r = self.c.fetchone()

        if r:
            return r[0]
        else:
            return None

    def executeFile(self, filename):
        sql = open(filename, 'r').read()

        if sql.strip():
            if self.scheme == 'sqlite':
                self.c.executescript(sql)

            elif self.scheme == 'pgsql':
                self.c.execute(sql)


    def execute(self, sql, data = None):
        if data:
            self.c.execute(sql, data)
        else:
            self.c.execute(sql)

    def begin(self):
        self.c.execute('BEGIN')

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()

    def close(self):
        self.c.close()
        self.conn.close()


#!/usr/bin/python
import argparse
import sys
from lib.main import Migranto

def main():

    parser = argparse.ArgumentParser(description='Simple SQL migration tool')

    parser.add_argument('--database', '-d', dest='database', metavar='URL', required=True,
                        help='database url (pgsql://user:password@host:port/dbname or sqlite://path/to/data.base)')

    parser.add_argument('--name', '-n', dest='name',
                        help='migration set name')

    parser.add_argument('--path', '-p', dest='path',
                        help='path to migrations dir')

    parser.add_argument('--migration', '-m', dest='migration', type=int, metavar = 'N',
                        help='migration step (last migration if omitted, 0 - before first migration)')

    parser.add_argument('--verbose', '-v', dest='verbose', action='count', default=False,
                        help='verbose mode')

    parser.add_argument('--storage', '-s', dest='storage', metavar='NAME', default='migranto',
                        help='migranto table name for data (default is migranto)')

    parser.add_argument('--fake', '-f', dest='fake', action='count', default=False,
                        help='fake migration (no sql applied to database, except migration number)')

    migranto = Migranto(parser.parse_args())

    try:
        migranto.run()
    except Exception, exc:
        print exc
        sys.exit(0)


sqlite = {
    'createTable': """
            CREATE TABLE {tablename}
            (
              "name" text PRIMARY KEY,
               migration int,
              "timestamp" timestamp not null
            );
        """,
    'insertMigration': "INSERT INTO {tablename} (name, migration, timestamp) VALUES (?, 0, time('now') );",
    'updateMigration': "UPDATE {tablename} SET migration = ? WHERE name=?;",
    'selectLatestMigration': "SELECT migration FROM {tablename} WHERE name=?;",
    'selectAllMigrationStatus': "SELECT * FROM {tablename};",
    'selectOneMigrationStatus': "SELECT * FROM {tablename} WHERE name=?;",
}

pgsql = {
    'createTable': """
                CREATE TABLE {tablename}
                (
                  "name" text PRIMARY KEY,
                   migration int,
                  "timestamp" timestamp not null
                );
            """,
    'insertMigration': "INSERT INTO {tablename} (name, migration, timestamp) VALUES (%s, 0, NOW() );",
    'updateMigration': "UPDATE {tablename} SET migration=%s WHERE name=%s;",
    'selectLatestMigration': "SELECT migration FROM {tablename} WHERE name=%s;",
    'selectAllMigrationStatus': "SELECT * FROM {tablename};",
    'selectOneMigrationStatus': "SELECT * FROM {tablename} WHERE name=%s;",
}


import glob
import os

def getFileListByMask(directory, masks):
    file_list = []
    for mask in masks:
        file_list.extend(glob.glob(os.path.join(directory, mask)))
        file_list.extend(glob.glob(os.path.join(directory, mask.upper())))
    return file_list


class objDict(dict):
    def __getattribute__(self, key):
        if key in self:
            return self[key]


import os
from setuptools import setup, find_packages

def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()

setup(
    name='migranto',
    version='0.1.7',
    author='Alexey Novikov',
    author_email='velocityzen@gmail.com',
    packages=find_packages(),
    install_requires = ["psycopg2"],
    url='https://github.com/velocityzen/migranto',
    license='BSD',
    description='Simple SQL migration tool for SQLite and PostgreSQL',
    long_description = read('README.md'),
    zip_safe=False,
    entry_points="""
    [console_scripts]
    migranto = migranto:main
    """
)



import datetime

class Printer:
    def __init__(self, out, format):
        self.out = out
        self.format = format

    def formatCell(self, data):
        if type(data) is datetime.datetime:
            return data.strftime('%Y-%m-%d %H:%M')
        else:
            return str(data)

    def printRow(self, row):

        print >> self.out, row[0].ljust(self.format[0] + 1),
        print >> self.out, self.formatCell(row[1]).rjust(self.format[1]) + '   ',
        print >> self.out, self.formatCell(row[2]).ljust(self.format[2]),
        print >> self.out